//////////////////////////////////////////////////////////////////////
// (c) Janusz Ganczarski
// http://www.januszg.hg.pl
// JanuszG@enter.net.pl
//////////////////////////////////////////////////////////////////////

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include "shaders.h"

//////////////////////////////////////////////////////////////////////
// rozmiary bryy obcinania
//////////////////////////////////////////////////////////////////////
GLfloat left = -2.0f;
GLfloat right = 2.0f;
GLfloat bottom = -2.0f;
GLfloat top = 2.0f;
GLfloat near = 3.0f;
GLfloat far = 7.0f;

//////////////////////////////////////////////////////////////////////
// macierz rzutowania
//////////////////////////////////////////////////////////////////////
glm::mat4x4 projectionMatrix;

//////////////////////////////////////////////////////////////////////
// macierz modelu-widoku
//////////////////////////////////////////////////////////////////////
glm::mat4x4 modelViewMatrix;

//////////////////////////////////////////////////////////////////////
// wspczynniki skalowania obiektu
//////////////////////////////////////////////////////////////////////
GLfloat scale = 1.0f;

//////////////////////////////////////////////////////////////////////
// kty obrotu obiektu
//////////////////////////////////////////////////////////////////////
GLfloat rotateX = 0.0f;
GLfloat rotateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// przesunicie obiektu
//////////////////////////////////////////////////////////////////////
GLfloat translateX = 0.0f;
GLfloat translateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu programu
//////////////////////////////////////////////////////////////////////
GLuint program;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora z danymi tablicy
// wierzchokw - wsprzdnymi wierzchokw krawdzi szecianu
//////////////////////////////////////////////////////////////////////
GLuint vertexBuffer;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora z danymi tablicy
// indeksw wierzchokw obiektu
//////////////////////////////////////////////////////////////////////
GLuint indicesBuffer;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu tablic wierzchokw
//////////////////////////////////////////////////////////////////////
GLuint vertexArray;

//////////////////////////////////////////////////////////////////////
// wsprzdne wierzchokw krawdzi skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLfloat position[8*3] =
{
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f,
    -1.0f, -1.0f, 1.0f,
    1.0f, -1.0f, 1.0f,
    1.0f, 1.0f, -1.0f,
    -1.0f, 1.0f, -1.0f,
    -1.0f, -1.0f, -1.0f,
    1.0f, -1.0f, -1.0f
};

//////////////////////////////////////////////////////////////////////
// dane indeksw wierzchokw krawdzi skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLuint indices[12*2] =
{
    5, 1,
    1, 0,
    0, 4,
    4, 5,
    2, 3,
    3, 0,
    1, 2,
    7, 4,
    3, 7,
    2, 6,
    6, 7,
    5, 6
};

//////////////////////////////////////////////////////////////////////
// dane indeksw wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLuint triangleIndices[12*3] =
{
    5, 1, 0,
    5, 0, 4,
    2, 3, 0,
    2, 0, 1,
    7, 4, 0,
    7, 0, 3,
    3, 2, 6,
    3, 6, 7,
    1, 6, 2,
    1, 5, 6,
    4, 6, 5,
    4, 7, 6
};

//////////////////////////////////////////////////////////////////////
// numery indeksw poszczeglnych atrybutw wierzchokw
//////////////////////////////////////////////////////////////////////
#define POSITION 0

//////////////////////////////////////////////////////////////////////
// funkcja generujca scen 3D
//////////////////////////////////////////////////////////////////////
void DisplayScene()
{
    // macierz modelu-widoku = macierz jednostkowa
    modelViewMatrix = glm::mat4x4( 1.0 );

    // przesunicie ukadu wsprzdnych obiektu do rodka bryy obcinania
    modelViewMatrix = glm::translate( modelViewMatrix, glm::vec3( 0.0f, 0.0f, -(near+far)/2.0f ) );

    // skalowanie obiektu
    modelViewMatrix = glm::scale( modelViewMatrix, glm::vec3( scale, scale, scale ) );

    // przesunicie obiektu
    modelViewMatrix = glm::translate( modelViewMatrix, glm::vec3( translateX, translateY, 0.0f ) );

    // obroty obiektu
    modelViewMatrix = glm::rotate( modelViewMatrix, rotateX, glm::vec3( 1.0f, 0.0f, 0.0f ) );
    modelViewMatrix = glm::rotate( modelViewMatrix, rotateY, glm::vec3( 0.0f, 1.0f, 0.0f ) );

    // czyszczenie bufora koloru
    glClear( GL_COLOR_BUFFER_BIT );

    // wczenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // wczenie programu
    glUseProgram( program );

    // zaadowanie zmiennej jednorodnej - iloczynu macierzy modelu-widoku i rzutowania
    glm::mat4x4 modelViewProjectionMatrix = projectionMatrix * modelViewMatrix;
    glUniformMatrix4fv( glGetUniformLocation( program, "modelViewProjectionMatrix" ), 1, GL_FALSE, glm::value_ptr( modelViewProjectionMatrix ) );

    // narysowanie danych zawartych w tablicach wierzchokw
    glDrawElements( GL_LINES, 12 * 2, GL_UNSIGNED_INT, NULL );

    // wyczenie programu
    glUseProgram( 0 );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );
}

//////////////////////////////////////////////////////////////////////
// zmiana wielkoci okna
//////////////////////////////////////////////////////////////////////
void Reshape( int width, int height )
{
    // obszar renderingu - cae okno
    glViewport( 0, 0, width, height );

    // parametry bryy obcinania - rzutowanie perspektywiczne
    // wysoko okna wiksza od szerokoci okna
    if( width < height && width > 0 )
         projectionMatrix = glm::frustum( left, right, bottom*height/width, top*height/width, near, far );
    else
        // szeroko okna wiksza lub rwna wysokoci okna
        if (width >= height && height > 0)
            projectionMatrix = glm::frustum( left*width/height, right*width/height, bottom, top, near, far );
        else
            projectionMatrix = glm::frustum( left, right, bottom, top, near, far );
}

//////////////////////////////////////////////////////////////////////
// inicjalizacja staych elementw maszyny stanu OpenGL
//////////////////////////////////////////////////////////////////////
void InitScene()
{
    // kolor ta - zawarto bufora koloru
    glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );

    // wczytanie shaderw i przygotowanie obsugi programu
    program = glCreateProgram();
    glAttachShader( program, LoadShader( GL_VERTEX_SHADER, "selekcja_obiektu_vs.glsl" ) );
    glAttachShader( program, LoadShader( GL_FRAGMENT_SHADER, "selekcja_obiektu_fs.glsl" ) );
    LinkValidateProgram( program );

    // generowanie identyfikatora obiektu tablic wierzchokw
    glGenVertexArrays( 1, &vertexArray );

    // utworzenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // generowanie identyfikatora obiektu bufora
    glGenBuffers( 1, &vertexBuffer );

    // utworzenie obiektu bufora wierzchokw (VBO) i zaadowanie danych
    glGenBuffers( 1, &vertexBuffer );
    glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer );
    glBufferData( GL_ARRAY_BUFFER, sizeof( position ), position, GL_STATIC_DRAW );
    glVertexAttribPointer( POSITION, 3, GL_FLOAT, GL_FALSE, 0, NULL );

    // wczenie tablic wierzchokw
    glEnableVertexAttribArray( POSITION );

    // utworzenie obiektu bufora indeksw wierzchokw i zaadowanie danych
    glGenBuffers( 1, &indicesBuffer );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indicesBuffer );
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, sizeof( indices ), indices, GL_STATIC_DRAW );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );
}

//////////////////////////////////////////////////////////////////////
// usunicie obiektw OpenGL
//////////////////////////////////////////////////////////////////////
void DeleteScene()
{
    // porzdki
    glDeleteProgram( program );
    glDeleteBuffers( 1, &vertexBuffer );
    glDeleteBuffers( 1, &indicesBuffer );
    glDeleteVertexArrays( 1, &vertexArray );
}

//////////////////////////////////////////////////////////////////////
// test, czy odcinek opisany przez punkt pocztkowy O i wektor
// kierunkowy D przecina trjkt opisany wierzchokami V0, V1 i V2
//////////////////////////////////////////////////////////////////////
bool PickTriangle( const glm::vec3 &V0, const glm::vec3 &V1, const glm::vec3 &V2,
                   const glm::vec3 &O, const glm::vec3 &D )
{
    // obliczenie wektora normalnego paszczyzny
    glm::vec3 N( glm::cross( V1 - V0, V2 - V0 ) );

    // obliczenie wspczynnika rwnania prostej wyznaczonej
    // przez punkt pocztkowy i wektor kierunkowy; jest
    // to jednoczenie odlego punktu przecicia od pocztku
    // promienia; jednoczenie obliczamy stron wielokta
    const float frontFacing = -glm::dot( N, D );
    const float dist = (glm::dot( N, O ) - glm::dot( V0, N )) / frontFacing;

    // punkt przecicia ley przed pocztkiem promienia
    if( dist < 0.0f ) return false;

    // punkt przecicia prostej z paszczyzn
    const glm::vec3 P( O + (D * dist) );

    // obliczenie wsprzdnych barycentrycznych punktu przecicia
    // prostej z paszczyzn
    const glm::vec3 uvw = glm::inverse( glm::mat3x3( V0, V1, V2 ) ) * P;

    // test przecicia
    if( uvw[0] < 0.0f || uvw[0] > 1.0f || uvw[1] < 0.0f || uvw[0] + uvw[1] > 1.0f ) return false;
    return true;
}

//////////////////////////////////////////////////////////////////////
// sprawdzenie, czy promie wychodzcy ze wsprzdnych
// okna (winX, winY) przecina obiekt - szecian
//////////////////////////////////////////////////////////////////////
bool PickObject( int winX, int winY )
{
    // pobranie obszaru renderingu
    GLint viewportInt[4];
    glGetIntegerv( GL_VIEWPORT, viewportInt );
    const glm::vec4 viewport( viewportInt[0], viewportInt[1], viewportInt[2], viewportInt[3] );

    // obliczenie wsprzdnych pocztku i koca odcinka (promienia) we wsprzdnych obiektu
    const glm::vec3 objNear = glm::unProject( glm::vec3( winX, winY, 0.0f ), modelViewMatrix, projectionMatrix, viewport );
    const glm::vec3 objFar = glm::unProject( glm::vec3( winX, winY, 1.0f ), modelViewMatrix, projectionMatrix, viewport );

    // obliczenie punktu pocztkowego odcinka (O) i wektora kierunkowego (D)
    const glm::vec3 O( objNear );
    const glm::vec3 D( objFar - objNear );

    // test przecicia kadego trjkta skadajcego si na obiekt
    int triangle = 0;
    while( triangle < 12 )
    {
        // wsprzdne wierzchokw trjkta
        const glm::vec3 V0( position[3 * triangleIndices[triangle * 3 + 0] + 0],
                            position[3 * triangleIndices[triangle * 3 + 0] + 1],
                            position[3 * triangleIndices[triangle * 3 + 0] + 2] );
        const glm::vec3 V1( position[3 * triangleIndices[triangle * 3 + 1] + 0],
                            position[3 * triangleIndices[triangle * 3 + 1] + 1],
                            position[3 * triangleIndices[triangle * 3 + 1] + 2] );
        const glm::vec3 V2( position[3 * triangleIndices[triangle * 3 + 2] + 0],
                            position[3 * triangleIndices[triangle * 3 + 2] + 1],
                            position[3 * triangleIndices[triangle * 3 + 2] + 2] );

        // test przecicia trjkta
        if( PickTriangle( V0, V1, V2, O, D ) )
            return true;

        // nastpny trjkt
        triangle++;
    }

    // brak przecicia obiektu
    return false;
}

//////////////////////////////////////////////////////////////////////
// obliczenie wsprzdnych przesunicia obiektu w oparciu
// o poprzednie (prevButton) oraz biece (currButton) pooenie
// kursora myszy
//////////////////////////////////////////////////////////////////////
glm::vec2 TranslateObject( const glm::vec2 &prevButton, const glm::vec2 &currButton )
{
    // pobranie obszaru renderingu
    GLint viewportInt[4];
    glGetIntegerv( GL_VIEWPORT, viewportInt );
    const glm::vec4 viewport( viewportInt[0], viewportInt[1], viewportInt[2], viewportInt[3] );

    // pobranie wsprzdnych w przestrzeni obiektu
    // dla poprzedniego pooenia kursora myszy
    const glm::vec3 prev = glm::unProject( glm::vec3( prevButton, 0.5f ), modelViewMatrix, projectionMatrix, viewport );

    // pobranie wsprzdnych w przestrzeni obiektu
    // dla biecego pooenia kursora myszy
    const glm::vec3 curr = glm::unProject( glm::vec3( currButton, 0.5f ), modelViewMatrix, projectionMatrix, viewport );

    // obliczenie wsprzdnych wektora przesunicia obiektu
    return glm::vec2( curr[0] - prev[0], curr[1] - prev[1] );
}
