//////////////////////////////////////////////////////////////////////
// (c) Janusz Ganczarski
// http://www.januszg.hg.pl
// JanuszG@enter.net.pl
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include "shaders.h"
#include "text.h"

//////////////////////////////////////////////////////////////////////
// rozmiary bryy obcinania
//////////////////////////////////////////////////////////////////////
GLfloat left = -2.0f;
GLfloat right = 2.0f;
GLfloat bottom = -2.0f;
GLfloat top = 2.0f;
GLfloat near = 3.0f;
GLfloat far = 7.0f;

//////////////////////////////////////////////////////////////////////
// macierz rzutowania
//////////////////////////////////////////////////////////////////////
glm::mat4x4 projectionMatrix;

//////////////////////////////////////////////////////////////////////
// wspczynniki skalowania obiektu
//////////////////////////////////////////////////////////////////////
GLfloat scale = 1.0f;

//////////////////////////////////////////////////////////////////////
// kty obrotu obiektu
//////////////////////////////////////////////////////////////////////
GLfloat rotateX = 0.0f;
GLfloat rotateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// przesunicie obiektu
//////////////////////////////////////////////////////////////////////
GLfloat translateX = 0.0f;
GLfloat translateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// numeracja obiektw programu
//////////////////////////////////////////////////////////////////////
enum
{
    FLAT_SHADING,       // cieniowanie paskie
    SMOOTH_SHADING,     // cieniowanie gadkie
    PROGRAM_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw programu
//////////////////////////////////////////////////////////////////////
GLuint program[PROGRAM_SIZE];

//////////////////////////////////////////////////////////////////////
// numeracja obiektw bufora wierzchokw oraz
// numery indeksw poszczeglnych atrybutw wierzchokw
//////////////////////////////////////////////////////////////////////
enum
{
    POSITION,
    COLOR,
    VERTEX_BUFFER_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw bufora z danymi tablic wierzchokw
//////////////////////////////////////////////////////////////////////
GLuint vertexBuffer[VERTEX_BUFFER_SIZE];

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora z danymi tablicy
// indeksw wierzchokw obiektu
//////////////////////////////////////////////////////////////////////
GLuint indicesBuffer;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu tablic wierzchokw
//////////////////////////////////////////////////////////////////////
GLuint vertexArray;

//////////////////////////////////////////////////////////////////////
// wsprzdne wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLfloat position[8*3] =
{
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f,
    -1.0f, -1.0f, 1.0f,
    1.0f, -1.0f, 1.0f,
    1.0f, 1.0f, -1.0f,
    -1.0f, 1.0f, -1.0f,
    -1.0f, -1.0f, -1.0f,
    1.0f, -1.0f, -1.0f
};

//////////////////////////////////////////////////////////////////////
// skadowe kolorw wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
const GLfloat color[8*3] =
{
    1.0f, 1.0f, 1.0f,
    1.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 0.0f,
    0.0f, 1.0f, 1.0f,
    1.0f, 0.0f, 1.0f,
    1.0f, 0.0f, 0.0f,
    0.0f, 0.0f, 0.0f,
    0.0f, 0.0f, 1.0f
};

//////////////////////////////////////////////////////////////////////
// dane indeksw wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLuint indices[12*3] =
{
    5, 1, 0,
    5, 0, 4,
    2, 3, 0,
    2, 0, 1,
    7, 4, 0,
    7, 0, 3,
    3, 2, 6,
    3, 6, 7,
    1, 6, 2,
    1, 5, 6,
    4, 6, 5,
    4, 7, 6
};

//////////////////////////////////////////////////////////////////////
// wybr wierzchoka podstawowego
//////////////////////////////////////////////////////////////////////
GLenum provokingVertex = GL_LAST_VERTEX_CONVENTION;

//////////////////////////////////////////////////////////////////////
// interpolacja kolorw wierzchokw wczona/wyczona
//////////////////////////////////////////////////////////////////////
GLboolean shading = SMOOTH_SHADING;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora ramki
//////////////////////////////////////////////////////////////////////
GLuint frameBuffer;

//////////////////////////////////////////////////////////////////////
// rozmiary obszarw renderingu do tekstury
//////////////////////////////////////////////////////////////////////
const int FRAME_WIDTH = 4*256;
const int FRAME_HEIGHT = 4*256;

//////////////////////////////////////////////////////////////////////
// numeracja obiektw tekstury
//////////////////////////////////////////////////////////////////////
enum
{
    COLOR0,
    DEPTH,
    RENDER_TEXTURE_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw tekstury
//////////////////////////////////////////////////////////////////////
GLuint renderTexture[RENDER_TEXTURE_SIZE];

//////////////////////////////////////////////////////////////////////
// rozmiary obszaru renderingu okna programu
//////////////////////////////////////////////////////////////////////
GLuint windowWidth, windowHeight;

//////////////////////////////////////////////////////////////////////
// funkcja generujca scen 3D
//////////////////////////////////////////////////////////////////////
void DisplayScene()
{
    // bufor ramki do zapisu = obiekt bufora ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, frameBuffer );

    // macierz modelu-widoku = macierz jednostkowa
    glm::mat4x4 modelViewMatrix = glm::mat4x4( 1.0 );

    // przesunicie ukadu wsprzdnych obiektu do rodka bryy obcinania
    modelViewMatrix = glm::translate( modelViewMatrix, glm::vec3( 0.0f, 0.0f, -(near+far)/2.0f ) );

    // skalowanie obiektu
    modelViewMatrix = glm::scale( modelViewMatrix, glm::vec3( scale, scale, scale ) );

    // przesunicie obiektu
    modelViewMatrix = glm::translate( modelViewMatrix, glm::vec3( translateX, translateY, 0.0f ) );

    // obroty obiektu
    modelViewMatrix = glm::rotate( modelViewMatrix, rotateX, glm::vec3( 1.0f, 0.0f, 0.0f ) );
    modelViewMatrix = glm::rotate( modelViewMatrix, rotateY, glm::vec3( 0.0f, 1.0f, 0.0f ) );

    // czyszczenie bufora koloru i bufora gbokoci
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // wybr wierzchoka podstawowego
    glProvokingVertex( provokingVertex );

    // wczenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // wczenie programu
    glUseProgram( program[shading] );

    // zaadowanie zmiennej jednorodnej - iloczynu macierzy modelu-widoku i rzutowania
    glm::mat4x4 modelViewProjectionMatrix = projectionMatrix * modelViewMatrix;
    glUniformMatrix4fv( glGetUniformLocation( program[shading], "modelViewProjectionMatrix" ), 1, GL_FALSE, glm::value_ptr( modelViewProjectionMatrix ) );

    // narysowanie danych zawartych w tablicach wierzchokw
    glDrawElements( GL_TRIANGLES, 12 * 3, GL_UNSIGNED_INT, NULL );

    // wyczenie programu
    glUseProgram( 0 );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );

    // bufor ramki do odczytu = obiekt bufora ramki
    glBindFramebuffer( GL_READ_FRAMEBUFFER, frameBuffer );

    // bufor ramki do zapisu = domylny bufor ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, 0 );

    // kopiowanie buforw
    glBlitFramebuffer( 0, 0, FRAME_WIDTH, FRAME_HEIGHT,
                       0, 0, windowWidth, windowHeight,
                       GL_COLOR_BUFFER_BIT, GL_LINEAR );

    // wyczenie tesktu gbokoci
    glDisable( GL_DEPTH_TEST );

    // wypisanie informacji o wierzchoku podstawowym
    DrawText8x16( 3, 3, provokingVertex == GL_FIRST_VERTEX_CONVENTION ?
                        "Wierzchoek podstawowy: GL_FIRST_VERTEX_CONVENTION" :
                        "Wierzchoek podstawowy: GL_LAST_VERTEX_CONVENTION" );

    // wypisanie informacji o rodzaju cieniowania
    DrawText8x16( 3, 21, shading == SMOOTH_SHADING ?
                         "Cieniowanie gadkie (smooth)" :
                         "Cieniowanie paskie (flat)" );

    // wczenie testu gbokoci
    glEnable( GL_DEPTH_TEST );
}

//////////////////////////////////////////////////////////////////////
// zmiana wielkoci okna
//////////////////////////////////////////////////////////////////////
void Reshape( int width, int height )
{
    // obszar renderingu - cae okno
    glViewportIndexedf( 0, 0.0, 0.0, static_cast<GLfloat>( width ), static_cast<GLfloat>( height ) );

    // zapisanie rozmiarw obszaru okna renderingu
    windowWidth = width;
    windowHeight = height;

    // parametry bryy obcinania - rzutowanie perspektywiczne
    // wysoko okna wiksza od szerokoci okna
    if( width < height && width > 0 )
         projectionMatrix = glm::frustum( left, right, bottom*height/width, top*height/width, near, far );
    else
        // szeroko okna wiksza lub rwna wysokoci okna
        if (width >= height && height > 0)
            projectionMatrix = glm::frustum( left*width/height, right*width/height, bottom, top, near, far );
        else
            projectionMatrix = glm::frustum( left, right, bottom, top, near, far );
}

//////////////////////////////////////////////////////////////////////
// inicjalizacja staych elementw maszyny stanu OpenGL
//////////////////////////////////////////////////////////////////////
void InitScene()
{
    // kolor ta - zawarto bufora koloru
    glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );

    // utworzenie obiektw tekstur - na dane bufora koloru i bufora gbokoci
    glGenTextures( RENDER_TEXTURE_SIZE, renderTexture );
    glBindTexture( GL_TEXTURE_2D, renderTexture[COLOR0] );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, FRAME_WIDTH, FRAME_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glBindTexture( GL_TEXTURE_2D, renderTexture[DEPTH] );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, FRAME_WIDTH, FRAME_HEIGHT, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

    // utworzenie obiektu bufora ramki i doczenie dwch obiektw bufora renderingu
    glGenFramebuffers( 1, &frameBuffer );
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, frameBuffer );
    glFramebufferTexture( GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, renderTexture[COLOR0], 0 );
    glFramebufferTexture( GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, renderTexture[DEPTH], 0 );

    // sprawdzenie poprawnoci i kompletnoci obiektu bufora ramki
    GLenum error = glCheckFramebufferStatus( GL_DRAW_FRAMEBUFFER );
    if( error != GL_FRAMEBUFFER_COMPLETE )
    {
        std::cout << "Niepoprawny obiekt bufora ramki" << std::endl;
        exit( 0 );
    }

    // obszar renderingu - caa powierzchnia bufora ramki
    glViewportIndexedf( 1, 0.0, 0.0, static_cast<GLfloat>( FRAME_WIDTH ), static_cast<GLfloat>( FRAME_HEIGHT ) );

    // wczytanie shaderw i przygotowanie obsugi programu
    program[FLAT_SHADING] = glCreateProgram();
    GLuint vertShader = LoadShader( GL_VERTEX_SHADER, "rendering_do_tekstury_vs.glsl" );
    glAttachShader( program[FLAT_SHADING], vertShader );
    glAttachShader( program[FLAT_SHADING], LoadShader( GL_GEOMETRY_SHADER, "rendering_do_tekstury_flat_gs.glsl" ) );
    glAttachShader( program[FLAT_SHADING], LoadShader( GL_FRAGMENT_SHADER, "rendering_do_tekstury_flat_fs.glsl" ) );
    LinkValidateProgram( program[FLAT_SHADING] );

    // wczytanie shaderw i przygotowanie obsugi programu
    program[SMOOTH_SHADING] = glCreateProgram();
    glAttachShader( program[SMOOTH_SHADING], vertShader );
    glAttachShader( program[SMOOTH_SHADING], LoadShader( GL_GEOMETRY_SHADER, "rendering_do_tekstury_smooth_gs.glsl" ) );
    glAttachShader( program[SMOOTH_SHADING], LoadShader( GL_FRAGMENT_SHADER, "rendering_do_tekstury_smooth_fs.glsl" ) );
    LinkValidateProgram( program[SMOOTH_SHADING] );

    // generowanie identyfikatora obiektu tablic wierzchokw
    glGenVertexArrays( 1, &vertexArray );

    // utworzenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // utworzenie obiektu bufora wierzchokw (VBO) i zaadowanie danych
    glGenBuffers( 1, &vertexBuffer[POSITION] );
    glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer[POSITION] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( position ), position, GL_STATIC_DRAW );
    glVertexAttribPointer( POSITION, 3, GL_FLOAT, GL_FALSE, 0, NULL );

    // utworzenie obiektu bufora wierzchokw (VBO) i zaadowanie danych
    glGenBuffers( 1, &vertexBuffer[COLOR] );
    glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer[COLOR] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( color ), color, GL_STATIC_DRAW );
    glVertexAttribPointer( COLOR, 3, GL_FLOAT, GL_FALSE, 0, NULL );

    // wczenie tablic wierzchokw
    glEnableVertexAttribArray( POSITION );
    glEnableVertexAttribArray( COLOR );

    // utworzenie obiektu bufora indeksw wierzchokw i zaadowanie danych
    glGenBuffers( 1, &indicesBuffer );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indicesBuffer );
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, sizeof( indices ), indices, GL_STATIC_DRAW );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );

    // wczenie testu bufora gbokoci
    glEnable( GL_DEPTH_TEST );

    // wczenie mechanizmw uywanych podczas renderingu tekstu
    InitDrawText();
}

//////////////////////////////////////////////////////////////////////
// usunicie obiektw OpenGL
//////////////////////////////////////////////////////////////////////
void DeleteScene()
{
    // porzdki
    glDeleteProgram( program[FLAT_SHADING] );
    glDeleteProgram( program[SMOOTH_SHADING] );
    glDeleteBuffers( VERTEX_BUFFER_SIZE, vertexBuffer );
    glDeleteBuffers( 1, &indicesBuffer );
    glDeleteVertexArrays( 1, &vertexArray );
    glDeleteTextures( RENDER_TEXTURE_SIZE, renderTexture );
    glDeleteFramebuffers( 1, &frameBuffer );

    // usunicie mechanizmw uywanych podczas renderingu tekstu
    DeleteDrawText();
}
