//////////////////////////////////////////////////////////////////////
// (c) Janusz Ganczarski
// http://www.januszg.hg.pl
// JanuszG@enter.net.pl
//////////////////////////////////////////////////////////////////////

#include <string>
#include <sstream>
#include <iostream>
#include <cstdlib>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include "shaders.h"
#include "text.h"
#include "textures.h"
#include "models.h"
#include "FreeImage.h"

//////////////////////////////////////////////////////////////////////
// nazwa pliku podana w wierszu polecenia
//////////////////////////////////////////////////////////////////////
char *fileName = NULL;

//////////////////////////////////////////////////////////////////////
// rozmiary bryy obcinania
//////////////////////////////////////////////////////////////////////
GLfloat left = -2.0f;
GLfloat right = 2.0f;
GLfloat bottom = -2.0f;
GLfloat top = 2.0f;
GLfloat near = 4.0f;
GLfloat far = 100.0f;

//////////////////////////////////////////////////////////////////////
// macierz rzutowania
//////////////////////////////////////////////////////////////////////
glm::mat4x4 projectionMatrix;

//////////////////////////////////////////////////////////////////////
// wspczynniki skalowania obiektu
//////////////////////////////////////////////////////////////////////
GLfloat scale = 1.0f;

//////////////////////////////////////////////////////////////////////
// kty obrotu skybox'a
//////////////////////////////////////////////////////////////////////
GLfloat rotateX = 0.0f;
GLfloat rotateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// przesunicie skybox'a
//////////////////////////////////////////////////////////////////////
GLfloat translateX = 0.0f;
GLfloat translateZ = 0.0f;

//////////////////////////////////////////////////////////////////////
// kty obrotu obiektu
//////////////////////////////////////////////////////////////////////
GLfloat modelRotateX = 0.0f;
GLfloat modelRotateY = 0.0f;

//////////////////////////////////////////////////////////////////////
// numeracja obiektw programu
//////////////////////////////////////////////////////////////////////
enum
{
    SKYBOX,
    OBJECT_3D,
    ENVIRONMENT_MAP,
    PROGRAM_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw programu
//////////////////////////////////////////////////////////////////////
GLuint program[PROGRAM_SIZE];

//////////////////////////////////////////////////////////////////////
// numeracja obiektw bufora wierzchokw
//////////////////////////////////////////////////////////////////////
enum
{
    SKYBOX_POSITION,
    SKYBOX_TEX_COORD,
    VERTEX_BUFFER_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw bufora z danymi tablic wierzchokw
//////////////////////////////////////////////////////////////////////
GLuint vertexBuffer[VERTEX_BUFFER_SIZE];

//////////////////////////////////////////////////////////////////////
// identyfikator obiektw tablic wierzchokw
//////////////////////////////////////////////////////////////////////
GLuint vertexArray;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora z danymi tablic
// indeksw wierzchokw obiektw
//////////////////////////////////////////////////////////////////////
GLuint indicesBuffer;

//////////////////////////////////////////////////////////////////////
// wsprzdne wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLfloat position[8*3] =
{
    20.0f, 20.0f, 20.0f,
    -20.0f, 20.0f, 20.0f,
    -20.0f, -20.0f, 20.0f,
    20.0f, -20.0f, 20.0f,
    20.0f, 20.0f, -20.0f,
    -20.0f, 20.0f, -20.0f,
    -20.0f, -20.0f, -20.0f,
    20.0f, -20.0f, -20.0f
};

//////////////////////////////////////////////////////////////////////
// skadowe wsprzdnych tekstury szeciennej
//////////////////////////////////////////////////////////////////////
const GLfloat texCoord[8*3] =
{
    1.0f, 1.0f, 1.0f,
    -1.0f, 1.0f, 1.0f,
    -1.0f, -1.0f, 1.0f,
    1.0f, -1.0f, 1.0f,
    1.0f, 1.0f, -1.0f,
    -1.0f, 1.0f, -1.0f,
    -1.0f, -1.0f, -1.0f,
    1.0f, -1.0f, -1.0f
};

//////////////////////////////////////////////////////////////////////
// dane indeksw wierzchokw trjktw skadajcych si na szecian
//////////////////////////////////////////////////////////////////////
GLuint indices[12*3] =
{
    5, 0, 1,
    5, 4, 0,
    2, 0, 3,
    2, 1, 0,
    7, 0, 4,
    7, 3, 0,
    3, 6, 2,
    3, 7, 6,
    1, 2, 6,
    1, 6, 5,
    4, 5, 6,
    4, 6, 7
};

//////////////////////////////////////////////////////////////////////
// numeracja obiektw tekstury
//////////////////////////////////////////////////////////////////////
enum
{
    CUBEMAP_MASKONAIVE,
    CUBEMAP_TEIDE,
    CUBEMAP_TENERIFE,
    CUBEMAP_VASA,
    TEXTURE_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw tekstury
//////////////////////////////////////////////////////////////////////
GLuint texture[TEXTURE_SIZE];

//////////////////////////////////////////////////////////////////////
// numer wybranego obiektu tekstury
//////////////////////////////////////////////////////////////////////
int texNumber = CUBEMAP_MASKONAIVE;

//////////////////////////////////////////////////////////////////////
// identyfikator obiektu bufora ramki
//////////////////////////////////////////////////////////////////////
GLuint frameBuffer;

//////////////////////////////////////////////////////////////////////
// rozmiary obszarw renderingu do tekstury
//////////////////////////////////////////////////////////////////////
const int FRAME_WIDTH = 4*256;
const int FRAME_HEIGHT = 4*256;

//////////////////////////////////////////////////////////////////////
// numeracja obiektw tekstury
//////////////////////////////////////////////////////////////////////
enum
{
    COLOR0,
    DEPTH,
    RENDER_TEXTURE_SIZE
};

//////////////////////////////////////////////////////////////////////
// identyfikatory obiektw tekstury
//////////////////////////////////////////////////////////////////////
GLuint renderTexture[RENDER_TEXTURE_SIZE];

//////////////////////////////////////////////////////////////////////
// numery indeksw poszczeglnych atrybutw wierzchokw
//////////////////////////////////////////////////////////////////////
#define POSITION  0
#define TEX_COORD 2

//////////////////////////////////////////////////////////////////////
// rodzaj modelu owietlenia Phonga/Blinna-Phonga/Lamberta
//////////////////////////////////////////////////////////////////////
enum
{
    PHONG_LIGHT,
    BLINN_PHONG_LIGHT,
    LAMBERT_LIGHT,
    PHONG_LIGHT_NORMAL_MAP,
    BLINN_PHONG_LIGHT_NORMAL_MAP,
    LAMBERT_LIGHT_NORMAL_MAP
};
int lightType = BLINN_PHONG_LIGHT_NORMAL_MAP;

//////////////////////////////////////////////////////////////////////
// znacznik odwrcenia wartoci mapy wysokoci
//////////////////////////////////////////////////////////////////////
bool reverseHeightMap = true;

//////////////////////////////////////////////////////////////////////
// obiekt klasy obsugujcej odczyt i rendering plikw 3D
//////////////////////////////////////////////////////////////////////
ModelAndTextureLoader model;

//////////////////////////////////////////////////////////////////////
// funkcja rysujca skybox'a
// modelViewMatrix - macierz modelu-widoku
//////////////////////////////////////////////////////////////////////
void DrawSkybox( glm::mat4x4 &modelViewMatrix )
{
    // bufor ramki do zapisu = domylny bufor ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, 0 );

    // czyszczenie bufora koloru i bufora gbokoci
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wybr obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, texture[texNumber] );

    // wczenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // wczenie programu
    glUseProgram( program[SKYBOX] );

    // zaadowanie zmiennej jednorodnej - iloczynu macierzy modelu-widoku i rzutowania
    glm::mat4x4 modelViewProjectionMatrix = projectionMatrix * modelViewMatrix;
    glUniformMatrix4fv( glGetUniformLocation( program[SKYBOX], "modelViewProjectionMatrix" ), 1, GL_FALSE, glm::value_ptr( modelViewProjectionMatrix ) );

    // zaadowanie zmiennej jednorodnej - numeru jednostki teksturujcej
    glUniform1i( glGetUniformLocation( program[SKYBOX], "tex" ), 6 );

    // narysowanie danych zawartych w tablicach wierzchokw
    glDrawElements( GL_TRIANGLES, 12 * 3, GL_UNSIGNED_INT, NULL );

    // wyczenie programu
    glUseProgram( 0 );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );

    // wyczenie obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );
}

//////////////////////////////////////////////////////////////////////
// rysowanie obiektu 3D
// objectModelViewMatrix - macierz modelu-widoku obiektu
// skyboxModelViewMatrix - macierz modelu-widoku skybox-a
//////////////////////////////////////////////////////////////////////
void DrawObject3D( glm::mat4x4 &objectModelViewMatrix, glm::mat4x4 skyboxModelViewMatrix )
{
    // bufor ramki do zapisu = domylny bufor ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, 0 );

    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wybr obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, renderTexture[COLOR0] );

    // odwrcona macierz modelu-widoku niezbdna do przeksztace
    // do ukadu wsprzdnych obiektu
    glm::mat4x4 modelViewMatrixInverse( glm::inverse( objectModelViewMatrix ) );

    // przeksztacenie wektora kierunku wiata do ukadu wsprzdnych obiektu
    glm::vec4 lightPosition( 0.0f, 0.0f, 1.0f, 0.0f );
    lightPosition = modelViewMatrixInverse * lightPosition;
    lightPosition = glm::normalize( lightPosition );

    // transformacja kierunku wiata do ukadu wsprzdnych obiektu
    glm::vec4 inverseLightPosition = modelViewMatrixInverse * lightPosition;
    inverseLightPosition = glm::normalize( inverseLightPosition );

    // przeksztacenie pooenia obserwatora do ukadu wsprzdnych obiektu
    glm::vec4 eyePosition( 0.0f, 0.0f, 0.0f, 1.0f );
    eyePosition = modelViewMatrixInverse * eyePosition;

    // wczenie programu
    glUseProgram( program[OBJECT_3D] );

    // zaadowanie zmiennej jednorodnej - iloczynu macierzy modelu-widoku i rzutowania
    glm::mat4x4 modelViewProjectionMatrix = projectionMatrix * objectModelViewMatrix;
    glUniformMatrix4fv( glGetUniformLocation( program[OBJECT_3D], "modelViewProjectionMatrix" ), 1, GL_FALSE, glm::value_ptr( modelViewProjectionMatrix ) );

    // zaadowanie zmiennej jednorodnej - macierzy modelu-widoku do oblicze odbicia
    glm::mat4x4 modelViewMatrix( glm::inverse( skyboxModelViewMatrix ) * objectModelViewMatrix );
    glUniformMatrix4fv( glGetUniformLocation( program[OBJECT_3D], "modelViewMatrix" ), 1, GL_FALSE, glm::value_ptr( modelViewMatrix ) );

    // zaadowanie zmiennej jednorodnej - macierzy przeksztacenia wektora normalnego
    // do oblicze odbicia; przeksztacenia macierzy wektora normalnego, tj. transponowanie
    // i odwracanie, nie s potrzebne z uwagi na to, e jest to macierz ortogonalna
    glm::mat3x3 normalMatrix( modelViewMatrix );
    glUniformMatrix3fv( glGetUniformLocation( program[OBJECT_3D], "normalMatrix" ), 1, GL_FALSE, glm::value_ptr( normalMatrix ) );

    // zaadowanie kierunku rda wiata i pooenia obserwatora w ukadzie wsprzdnych obiektu
    glUniform4fv( glGetUniformLocation( program[OBJECT_3D], "lightSource[0].position" ), 1, glm::value_ptr( lightPosition ) );
    glUniform4fv( glGetUniformLocation( program[OBJECT_3D], "eyePosition" ), 1, glm::value_ptr( eyePosition ) );

    // pobranie indeksu funkcji obsugujcej wybrany rodzaj modelu owietlenia
    const char *lightName[] = { "Phong", "BlinnPhong", "Lambert", "PhongNormalMap", "BlinnPhongNormalMap", "LambertNormalMap" };
    GLuint light = glGetSubroutineIndex( program[OBJECT_3D], GL_FRAGMENT_SHADER, lightName[lightType] );

    // zaadowanie zmiennej jednorodnej podprogramu - rodzaj modelu owietlenia
    glUniformSubroutinesuiv( GL_FRAGMENT_SHADER, 1, &light );

    // wspczynnik skalowania wielkoci przemieszczenia
    GLfloat displacement = 0.015f * model.GetSize();
    glUniform1f( glGetUniformLocation( program[OBJECT_3D], "displacement" ), displacement );

    // znacznik odwrcenia wartoci mapy wysokoci
    glUniform1i( glGetUniformLocation( program[OBJECT_3D], "reverseHeightMap" ), reverseHeightMap );

    // przeksztacenie wsprzdnych oka do ukadu wsprzdnych obserwatora
    glm::vec3 eye( 0.0f );
    eye += glm::vec3 ( translateX, 0.0f, translateZ );
    glUniform3fv( glGetUniformLocation( program[OBJECT_3D], "eye" ), 1, glm::value_ptr( eye ) );

    // obiekt zawiera materiay pprzezroczyste
    if( model.GetOpacity() )
    {
        // narysowanie danych obiektw nieprzezroczystych
        model.MeshRenderingOpacity( false );

        // wczenie mieszania kolorw
        glEnable( GL_BLEND );

        // wyczenie zapisu danych bufora gbokoci
        glDepthMask( GL_FALSE );

        // wczenie renderingu wybranej strony wielokta
        glEnable( GL_CULL_FACE );

        // rysowanie tylko tylnej strony wieloktw
        glCullFace( GL_FRONT );

        // narysowanie danych (p)przezroczystych
        model.MeshRenderingOpacity( true );

        // rysowanie tylko przedniej strony wieloktw
        glCullFace( GL_BACK );

        // narysowanie danych (p)przezroczystych
        model.MeshRenderingOpacity( true );

        // wyczenie renderingu wybranej strony wielokta
        glDisable( GL_CULL_FACE );

        // wczenie zapisu danych do bufora gbokoci
        glDepthMask( GL_TRUE );

        // wyczenie mieszania kolorw
        glDisable( GL_BLEND );
    }
    else
        // rendering obiektu 3D z pliku
        model.MeshRendering();

    // wyczenie programu
    glUseProgram( 0 );

    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wyczenie obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );
}

//////////////////////////////////////////////////////////////////////
// zapis tekstury kubicznej do plikw graficznych
//////////////////////////////////////////////////////////////////////
void SaveCubeMap()
{
    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wybr obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, renderTexture[COLOR0] );

    // nazwy zapisywanych plikw
    const char *fileNames[] =
    {
        "positive_x.png",
        "negative_x.png",
        "positive_y.png",
        "negative_y.png",
        "positive_z.png",
        "negative_z.png"
    };

    // zapis danych tekstury do 6-ciu plikw graficznych
    SaveCubeTexture( fileNames );

    // wyczenie obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );
}

//////////////////////////////////////////////////////////////////////
// generowanie tekstury kubicznej
// skyboxModelViewMatrix - macierz modelu-widoku
//////////////////////////////////////////////////////////////////////
void GenerateCubeMap( glm::mat4x4 &skyboxModelViewMatrix )
{
    // bufor ramki do zapisu = obiekt bufora ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, frameBuffer );

    // czyszczenie bufora koloru i bufora gbokoci
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wybr obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, texture[texNumber] );

    // wczenie obiektu tablic wierzchokw
    glBindVertexArray( vertexArray );

    // wczenie programu
    glUseProgram( program[ENVIRONMENT_MAP] );

    // zaadowanie zmiennej jednorodnej - uchwyt tekstury
    glUniform1i( glGetUniformLocation( program[ENVIRONMENT_MAP], "tex" ), 6 );

    // parametry bryy obcinania - kt widzenia wynosi 90
    glm::mat4x4 projectionMatrix =  glm::perspective( 90.0f, 1.0f, 1.0f, 100.0f );

    // wsprzdne punktw, w ktrych kierunku jest zwrcony obserwator
    const glm::vec3 center[6] =
    {
        glm::vec3(  1.0f,  0.0f,  0.0f ),
        glm::vec3( -1.0f,  0.0f,  0.0f ),
        glm::vec3(  0.0f,  1.0f,  0.0f ),
        glm::vec3(  0.0f, -1.0f,  0.0f ),
        glm::vec3(  0.0f,  0.0f,  1.0f ),
        glm::vec3(  0.0f,  0.0f, -1.0f )
    };

    // wsprzdne wektorw okrelajcych kierunek "do gry"
    const glm::vec3 up[6] =
    {
        glm::vec3( 0.0f, -1.0f,  0.0f ),
        glm::vec3( 0.0f, -1.0f,  0.0f ),
        glm::vec3( 0.0f,  0.0f,  1.0f ),
        glm::vec3( 0.0f,  0.0f, -1.0f ),
        glm::vec3( 0.0f, -1.0f,  0.0f ),
        glm::vec3( 0.0f, -1.0f,  0.0f ) 
    };

    // obliczenie i zaadowanie macierzy
    for( int face = 0; face < 6; face++ )
    {
        // obliczenie macierzy modelu-widoku biecej strony tekstury szeciennej
        glm::mat4x4 modelViewMatrix = glm::lookAt( glm::vec3( 0.0f, 0.0f, 0.0f ), center[face], up[face] );
        modelViewMatrix *= skyboxModelViewMatrix;

        // obliczenie i zaadowanie iloczynu macierzy rzutowania
        // i modelu-widoku biecej strony tekstury szeciennej
        glm::mat4x4 modelViewProjectionMatrix = projectionMatrix * modelViewMatrix;
        std::ostringstream txt;
        txt << "modelViewProjectionMatrix[" << face << "]";
        glUniformMatrix4fv( glGetUniformLocation( program[ENVIRONMENT_MAP], txt.str().c_str() ), 1, GL_FALSE, glm::value_ptr( modelViewProjectionMatrix ) );
    }

    // narysowanie danych zawartych w tablicach wierzchokw
    glDrawElements( GL_TRIANGLES, 12 * 3, GL_UNSIGNED_INT, NULL );

    // wyczenie programu
    glUseProgram( 0 );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );

    // wybr aktywnej jednostki teksturujcej
    glActiveTexture( GL_TEXTURE6 );

    // wyczenie obiektu tekstury
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );
}

//////////////////////////////////////////////////////////////////////
// funkcja generujca scen 3D
//////////////////////////////////////////////////////////////////////
void DisplayScene()
{
    // macierz modelu-widoku skybox'a - macierz jednostkowa
    glm::mat4x4 skyboxModelViewMatrix = glm::mat4x4( 1.0 );

    // przesunicie obserwatora do punktu pocztkowego wewntrz skybox'a
    skyboxModelViewMatrix = glm::translate( skyboxModelViewMatrix, glm::vec3( translateX, 0.0f, translateZ ) );

    // obroty skybox'a
    skyboxModelViewMatrix = glm::rotate( skyboxModelViewMatrix, rotateX, glm::vec3( 1.0f, 0.0f, 0.0f ) );
    skyboxModelViewMatrix = glm::rotate( skyboxModelViewMatrix, rotateY, glm::vec3( 0.0f, 1.0f, 0.0f ) );

    // macierz modelu-widoku obiektu = macierz jednostkowa
    glm::mat4x4 objectModelViewMatrix = glm::mat4x4( 1.0 );

    // przesunicie obserwatora tak, aby ukad wsprzdnych obiektu by w rodku bryy obcinania
    objectModelViewMatrix = glm::translate( objectModelViewMatrix, glm::vec3( 0.0f, 0.0f, -10.0f ) );

    // skalowanie obiektu do wielkoci bryy obcinania
    objectModelViewMatrix = glm::scale( objectModelViewMatrix, glm::vec3( (right - left) * 0.8f / model.GetSize() ) );

    // skalowanie obiektu
    objectModelViewMatrix = glm::scale( objectModelViewMatrix, glm::vec3( scale, scale, scale ) );

    // obroty obiektu
    objectModelViewMatrix = glm::rotate( objectModelViewMatrix, modelRotateX, glm::vec3( 1.0f, 0.0f, 0.0f ) );
    objectModelViewMatrix = glm::rotate( objectModelViewMatrix, modelRotateY, glm::vec3( 0.0f, 1.0f, 0.0f ) );

    // wycentowanie obiektu
    objectModelViewMatrix = glm::translate( objectModelViewMatrix, -model.GetCenter() );

    // narysowanie skybox'a
    DrawSkybox( skyboxModelViewMatrix );

    // generowanie mapy szeciennej z tekstur otoczenia (rodowiska)
    GenerateCubeMap( skyboxModelViewMatrix );

    // narysowanie obiektu
    DrawObject3D( objectModelViewMatrix, glm::mat4( 1.0 ) );

    // wypisanie rodzaju modelu owietlenia
    const char* ligtTypeStr[] = { "owietlenie Phonga",
                                  "owietlenie Blinna-Phonga",
                                  "owietlenie Lamberta",
                                  "owietlenie Phonga z mapowaniem wektorw normalnych",
                                  "owietlenie Blinna-Phonga z mapowaniem wektorw normalnych",
                                  "owietlenie Lamberta z mapowaniem wektorw normalnych" };
    DrawText8x16( 3, 3, ligtTypeStr[lightType], glm::vec4( 1.0f ) );
}

//////////////////////////////////////////////////////////////////////
// zmiana wielkoci okna
//////////////////////////////////////////////////////////////////////
void Reshape( int width, int height )
{
    // obszar renderingu - cae okno
    glViewportIndexedf( 0, 0.0, 0.0, static_cast<GLfloat>( width ), static_cast<GLfloat>( height ) );

    // parametry bryy obcinania - rzutowanie perspektywiczne
    // wysoko okna wiksza od szerokoci okna
    // wysoko okna wiksza od szerokoci okna
    if( width < height && width > 0 )
         projectionMatrix = glm::frustum( left, right, bottom*height/width, top*height/width, near, far );
    else
        // szeroko okna wiksza lub rwna wysokoci okna
        if (width >= height && height > 0)
            projectionMatrix = glm::frustum( left*width/height, right*width/height, bottom, top, near, far );
        else
            projectionMatrix = glm::frustum( left, right, bottom, top, near, far );
}

//////////////////////////////////////////////////////////////////////
// inicjalizacja staych elementw maszyny stanu OpenGL
//////////////////////////////////////////////////////////////////////
void InitScene()
{
    // kolor ta - zawarto bufora koloru
    glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );

    // zaadowanie danych z pliku 3D
    if( !model.Load( fileName ) )
    {
        std::cout << "Niepoprawny odczyt pliku " << fileName << std::endl;
        exit( 0 );
    }

    // utworzenie obiektu tekstury na dane bufora koloru
    glGenTextures( RENDER_TEXTURE_SIZE, renderTexture );
    glBindTexture( GL_TEXTURE_CUBE_MAP, renderTexture[COLOR0] );
    for( int face = 0; face < 6; face++ )
    {
        glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_RGBA, FRAME_WIDTH, FRAME_HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
    }
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

    // utworzenie obiektu tekstury na dane bufora gbokoci
    glBindTexture( GL_TEXTURE_CUBE_MAP, renderTexture[DEPTH] );
    for( int face = 0; face < 6; face++ )
    {
        glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 0, GL_DEPTH_COMPONENT, FRAME_WIDTH, FRAME_HEIGHT, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL );
    }
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

    // utworzenie obiektu bufora ramki i doczenie dwch obiektw tekstury
    glGenFramebuffers( 1, &frameBuffer );
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, frameBuffer );
    glFramebufferTexture( GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, renderTexture[COLOR0], 0 );
    glFramebufferTexture( GL_DRAW_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, renderTexture[DEPTH], 0 );

    // sprawdzenie poprawnoci i kompletnoci obiektu bufora ramki
    GLenum error = glCheckFramebufferStatus( GL_DRAW_FRAMEBUFFER );
    if( error != GL_FRAMEBUFFER_COMPLETE )
    {
        std::cout << "Niepoprawny obiekt bufora ramki" << std::endl;
        exit( 0 );
    }

    // obszar renderingu - caa powierzchnia bufora ramki
    glViewportIndexedf( 1, 0.0, 0.0, static_cast<GLfloat>( FRAME_WIDTH ), static_cast<GLfloat>( FRAME_HEIGHT ) );

    // bufor ramki do zapisu = domylny bufor ramki
    glBindFramebuffer( GL_DRAW_FRAMEBUFFER, 0 );

    // wczytanie shaderw i przygotowanie obsugi programu
    program[SKYBOX] = glCreateProgram();
    glAttachShader( program[SKYBOX], LoadShader( GL_VERTEX_SHADER, "skybox_vs.glsl" ) );
    glAttachShader( program[SKYBOX], LoadShader( GL_FRAGMENT_SHADER, "skybox_fs.glsl" ) );
    LinkValidateProgram( program[SKYBOX] );

    // wczytanie shaderw i przygotowanie obsugi programu
    program[OBJECT_3D] = glCreateProgram();
    glAttachShader( program[OBJECT_3D], LoadShader( GL_VERTEX_SHADER, "mapowanie_otoczenia_vs.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_TESS_CONTROL_SHADER, "mapowanie_otoczenia_tcs.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_TESS_EVALUATION_SHADER, "mapowanie_otoczenia_tes.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_FRAGMENT_SHADER, "../../common/light_model_static.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_FRAGMENT_SHADER, "../../common/blinn_phong_light.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_FRAGMENT_SHADER, "../../common/phong_light.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_FRAGMENT_SHADER, "../../common/lambert_light.glsl" ) );
    glAttachShader( program[OBJECT_3D], LoadShader( GL_FRAGMENT_SHADER, "mapowanie_otoczenia_fs.glsl" ) );
    LinkValidateProgram( program[OBJECT_3D] );

    // wczytanie shaderw i przygotowanie obsugi programu
    program[ENVIRONMENT_MAP] = glCreateProgram();
    glAttachShader( program[ENVIRONMENT_MAP], LoadShader( GL_VERTEX_SHADER, "generowanie_tekstury_szeciennej_vs.glsl" ) );
    glAttachShader( program[ENVIRONMENT_MAP], LoadShader( GL_GEOMETRY_SHADER, "generowanie_tekstury_szeciennej_gs.glsl" ) );
    glAttachShader( program[ENVIRONMENT_MAP], LoadShader( GL_FRAGMENT_SHADER, "skybox_fs.glsl" ) );
    LinkValidateProgram( program[ENVIRONMENT_MAP] );

    // utworzenie obiektu tablic wierzchokw
    glGenVertexArrays( 1, &vertexArray );
    glBindVertexArray( vertexArray );

    // utworzenie obiektu bufora wierzchokw (VBO) i zaadowanie danych
    glGenBuffers( 1, &vertexBuffer[SKYBOX_POSITION] );
    glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer[SKYBOX_POSITION] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( position ), position, GL_STATIC_DRAW );
    glVertexAttribPointer( POSITION, 3, GL_FLOAT, GL_FALSE, 0, NULL );

    // utworzenie obiektu bufora wierzchokw (VBO) i zaadowanie danych
    glGenBuffers( 1, &vertexBuffer[SKYBOX_TEX_COORD] );
    glBindBuffer( GL_ARRAY_BUFFER, vertexBuffer[SKYBOX_TEX_COORD] );
    glBufferData( GL_ARRAY_BUFFER, sizeof( texCoord ), texCoord, GL_STATIC_DRAW );
    glVertexAttribPointer( TEX_COORD, 3, GL_FLOAT, GL_FALSE, 0, NULL );

    // wczenie tablic wierzchokw
    glEnableVertexAttribArray( POSITION );
    glEnableVertexAttribArray( TEX_COORD );

    // utworzenie obiektu bufora indeksw wierzchokw i zaadowanie danych
    glGenBuffers( 1, &indicesBuffer );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, indicesBuffer );
    glBufferData( GL_ELEMENT_ARRAY_BUFFER, sizeof( indices ), indices, GL_STATIC_DRAW );

    // wyczenie obiektu tablic wierzchokw
    glBindVertexArray( 0 );

    // nazwy plikw z obrazami trzech tekstur szeciennych
    // dane pochodz ze strony WWW: http://www.humus.name
    const std::string fileNames[6*TEXTURE_SIZE] =
    {
        "../../media/www.humus.name/maskonaive/posx.jpg",
        "../../media/www.humus.name/maskonaive/negx.jpg",
        "../../media/www.humus.name/maskonaive/posy.jpg",
        "../../media/www.humus.name/maskonaive/negy.jpg",
        "../../media/www.humus.name/maskonaive/posz.jpg",
        "../../media/www.humus.name/maskonaive/negz.jpg",

        "../../media/www.humus.name/teide/posx.jpg",
        "../../media/www.humus.name/teide/negx.jpg",
        "../../media/www.humus.name/teide/posy.jpg",
        "../../media/www.humus.name/teide/negy.jpg",
        "../../media/www.humus.name/teide/posz.jpg",
        "../../media/www.humus.name/teide/negz.jpg",

        "../../media/www.humus.name/tenerife/posx.jpg",
        "../../media/www.humus.name/tenerife/negx.jpg",
        "../../media/www.humus.name/tenerife/posy.jpg",
        "../../media/www.humus.name/tenerife/negy.jpg",
        "../../media/www.humus.name/tenerife/posz.jpg",
        "../../media/www.humus.name/tenerife/negz.jpg",

        "../../media/www.humus.name/vasa/posx.jpg",
        "../../media/www.humus.name/vasa/negx.jpg",
        "../../media/www.humus.name/vasa/posy.jpg",
        "../../media/www.humus.name/vasa/negy.jpg",
        "../../media/www.humus.name/vasa/posz.jpg",
        "../../media/www.humus.name/vasa/negz.jpg"
    };

    // kolejne obrazy tekstury szeciennej
    const GLenum targets[6] =
    {
        GL_TEXTURE_CUBE_MAP_POSITIVE_X,
        GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
        GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
        GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
        GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
        GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
    };

    // generowanie trzech identyfikatorw obiektw tekstury
    glGenTextures( TEXTURE_SIZE, texture );

    // utworzenie trzech tekstur szeciennych
    for( int i = 0; i < TEXTURE_SIZE; i++ )
    {
        // utworzenie obiektu tekstury
        glBindTexture( GL_TEXTURE_CUBE_MAP, texture[i] );

        // odczyt kolejnych plikw z tekstur i ewentualny komunikat o bdzie
        for( int j = 0; j < 6; j++ )
            if( !LoadTexture( fileNames[i*6+j].c_str(), targets[j] ) )
            {
                std::cout << "Niepoprawny odczyt pliku " << fileNames[i*6+j] << std::endl;
                exit( 0 );
            }

        // filtr powikszajcy i pomniejszajcy
        glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

        // tryb zawijania
        glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
        glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
        glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
    }

    // przeczenie na tekstur domyln
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0 );

    // bezszwowa filtracja tekstury szeciennej
    glEnable( GL_TEXTURE_CUBE_MAP_SEAMLESS );

    // wczenie testu bufora gbokoci
    glEnable( GL_DEPTH_TEST );

    // wczenie mechanizmw uywanych podczas renderingu tekstu
    InitDrawText();

    // wczenie ustawie uywanych podczas renderingu obiektu pprzezroczystego
    if( model.GetOpacity() )
    {
        // wspczynniki mieszania kolorw
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    }
}

//////////////////////////////////////////////////////////////////////
// usunicie obiektw OpenGL
//////////////////////////////////////////////////////////////////////
void DeleteScene()
{
    // porzdki
    glDeleteProgram( program[SKYBOX] );
    glDeleteProgram( program[OBJECT_3D] );
    glDeleteProgram( program[ENVIRONMENT_MAP] );
    glDeleteBuffers( VERTEX_BUFFER_SIZE, vertexBuffer );
    glDeleteVertexArrays( 1, &vertexArray );
    glDeleteBuffers( 1, &indicesBuffer );
    glDeleteTextures( TEXTURE_SIZE, texture );

    // usunicie mechanizmw uywanych podczas renderingu tekstu
    DeleteDrawText();
}
